/*********************** 2-D MODULE BUCHBERGER ROUTINES ***********************/

/******************************************************************************/
#include "buch2.h"
#include "meataxe.h"

static dclass_t *doubleInnerNewClass(void)
{
  register dclass_t *d = (dclass_t *) malloc(sizeof(dclass_t));
  if (!d) AllocationError("doubleInnerNewClass");
  d->c[0] = NULL;
  d->c[1] = NULL;
  return d;
}

/******************************************************************************/
dclass_t *doubleCompletelyNewClass(void)
{
  register dclass_t *d = doubleInnerNewClass();
  d->c[0] = newClass();
  d->c[1] = newClass();
  return d;
}

/******************************************************************************/
dclass_t *doubleNewClass(cclass_t *c1, cclass_t *c2)
{
  register dclass_t *d = doubleInnerNewClass();
  d->c[0] = c1;
  d->c[1] = c2;
  return d;
}

/******************************************************************************/
static void doubleFreeClass(dclass_t *d)
{
  if (d->c[0]) freeClass(d->c[0]);
  if (d->c[1]) freeClass(d->c[1]);
  free(d);
  return;
}

/******************************************************************************/
static void doubleWriteOutClass(FILE *fp, freeGcr_t *fGcr, dclass_t *d)
{
  fprintf(fp, "Sum1: ");
  writeOutClass(fp, fGcr, d->c[0]);
  fprintf(fp, "Sum2: ");
  writeOutClass(fp, fGcr, d->c[1]);
  return;
}

/******************************************************************************/
void doublePrintClass(freeGcr_t *fGcr, dclass_t *d)
{
  doubleWriteOutClass(stdout, fGcr, d);
  return;
}

/******************************************************************************/
drel_t *doubleNewRel(cterm_t *tip, dclass_t *val, boolean decom)
{
  register drel_t *r = (drel_t *) malloc(sizeof(drel_t));
  if (!r) AllocationError("doubleNewRel");
  r->tip = tip;
  r->val = val;
  r->decom = decom;
  r->prev = NULL;
  r->next = NULL;
  return r;
}

/******************************************************************************/
void doubleFreeRel(drel_t *r)
{
  if (r->tip) freeTerm(r->tip);
  if (r->val) doubleFreeClass(r->val);
  free(r);
  return;
}

/******************************************************************************/
dunred_t *doubleNewUnred(dclass_t *d, boolean decom)
{
  register dunred_t *u = (dunred_t *) malloc(sizeof(dunred_t));
  if (!u) AllocationError("doubleNewUnred");
  u->d = d;
  u->decom = decom;
  u->prev = NULL;
  u->next = NULL;
  return u;
}

/******************************************************************************/
static void doubleFreeUnred(dunred_t *u)
{
  if (u->d) doubleFreeClass(u->d);
  free(u);
  return;
}

/******************************************************************************/
static dpair_t *doubleNewCritPair(void)
{
  register dpair_t *cp = (dpair_t *) malloc(sizeof(dpair_t));
  if (!cp) AllocationError("doubleNewCritPair");
  cp->lcm = NULL;
  cp->prev = NULL;
  cp->next = NULL;
  cp->r1 = NULL;
  cp->t1 = NULL;
  cp->r2 = NULL;
  cp->t2 = NULL;
  return cp;
}

/******************************************************************************/
static void doubleFreeCritPair(dpair_t *cp)
{
  if (cp->lcm) freeTerm(cp->lcm);
  if (cp->t1) freeTerm(cp->t1);
  if (cp->t2) freeTerm(cp->t2);
  free(cp);
  return;
}

/******************************************************************************/
benef_t *newBenef(dideal_t *di, long parity)
{
  register benef_t *benef = (benef_t *) malloc(sizeof(benef_t));
  if (!benef) AllocationError("newBenef");
  benef->di = di;
  benef->parity = parity;
  return benef;
}

/******************************************************************************/
void freeBenef(benef_t *benef)
{
  free(benef);
  return;
}

/******************************************************************************/
static boolean LTinFirstSummand(dclass_t *d)
{
  return (d->c[0]->firstMon) ? true : false;
}

/******************************************************************************/
static boolean LTinSecondSummand(dclass_t *d)
{
  if (LTinFirstSummand(d)) return false;
  return (d->c[1]->firstMon) ? true : false;
}

/******************************************************************************/
void doubleInsertUnredRel(dideal_t *di, dunred_t *u)
{
  register dunred_t *next, *prev = NULL;
  register dclass_t *d = u->d;
  if (!LTinFirstSummand(d))
    OtherError("doubleInsertUnredRel: LT not in first summand");
  for (next = di->firstUnredRel;
    next && leqTerm(d->c[0]->firstMon->term, next->d->c[0]->firstMon->term);
    next = next->next) prev = next;
  u->prev = prev;
  u->next = next;
  if (prev)
    prev->next = u;
  else
    di->firstUnredRel = u;
  if (next) next->prev = u;
  return;
}

/******************************************************************************/
static long doubleDimUnred(dideal_t *di, dunred_t *u)
{
  register dclass_t *d = u->d;
  cmon_t *m;
  if ((m = d->c[0]->firstMon)) return m->term->dim;
  if ((m = d->c[1]->firstMon)) return m->term->dim + di->diff;
  OtherError("doubleDimUnred: class is zero");
  return 0; /* Of course, this line will never be executed */
}

/******************************************************************************/
static long doubleInnerSmallestInterestingDimension(dideal_t *di)
{
  register long d = -1, n, d2;
  if (di->firstUnredRel)
    d = doubleDimUnred(di, di->firstUnredRel);
  if (di->firstCritPair)
  {
    n = di->firstCritPair->lcm->dim;
    d = modifiedMinlong(d, n);
  }
  d2 = smallestInterestingDimension(di->secondSummand);
  if (d2 != -1) d = modifiedMinlong(d, d2 + di->diff);
  return d;
}

/******************************************************************************/
void doubleBuchberger(dideal_t *di);

/******************************************************************************/
static void doubleReactToAnyNewExteriorGenerators(dideal_t *di)
{
  register freeGcr_t *fGcr;
  register long i;
  boolean shouldDoBuchberger;
  fGcr = di->fGcr;
  reactToAnyNewExteriorGenerators(di->secondSummand);
  if (!isOdd(fGcr->p) || di->nExtGens == fGcr->nExtGens) return;
  if (di->nExtGens > fGcr->nExtGens)
    OtherError("doubleReactToAnyNewExteriorGenerators: corrupt data");
  if (!di->newExtGensAllowed)
    OtherError("doubleReactToAnyNewExteriorGenerators: prohibited");
  shouldDoBuchberger = (doubleInnerSmallestInterestingDimension(di) == -1) ?
    true : false;
  for (i = di->nExtGens; i < fGcr->nExtGens; i++)
  {
    register long e;
    register cclass_t *c1, *c2;
    register dclass_t *d;
    register dunred_t *u;
    e = fGcr->numExtGen[i];
    c1 = squareGenClass(fGcr, e);
    c2 = newClass();
    d = doubleNewClass(c1, c2);
    u = doubleNewUnred(d, false);
    doubleInsertUnredRel(di, u);
  }
  di->nExtGens = fGcr->nExtGens;
  if (shouldDoBuchberger) doubleBuchberger(di);
  return;
}

/******************************************************************************/
long doubleSmallestInterestingDimension(dideal_t *di)
{
  doubleReactToAnyNewExteriorGenerators(di);
  return doubleInnerSmallestInterestingDimension(di);
}

/******************************************************************************/
static dideal_t *doubleInnerNewHomogIdeal(void)
{
  register dideal_t *di = (dideal_t *) malloc(sizeof(dideal_t));
  if (!di) AllocationError("doubleInnerNewHomogIdeal");
  di->fGcr = NULL;
  di->firstRedRel = NULL;
  di->firstUnredRel = NULL;
  di->firstCritPair = NULL;
  di->secondSummand = NULL;
  di->diff = 0;
  di->nExtGens = 0;
  di->newExtGensAllowed = true;
  return di;
}

/******************************************************************************/
dideal_t *doubleNewHomogIdeal(freeGcr_t *fGcr, long diff)
{
  register dideal_t *di = doubleInnerNewHomogIdeal();
  if (diff < 0) OtherError("doubleNewHomogIdeal: diff >= 0 assumed");
  di->fGcr = fGcr;
  di->diff = diff;
  di->secondSummand = newHomogIdeal(fGcr);
  di->benef = NULL;
  doubleReactToAnyNewExteriorGenerators(di);
  return di;
}

/******************************************************************************/
void doubleFreeHomogIdeal(dideal_t *di)
{
  register dunred_t *u;
  register drel_t *rel;
  register dpair_t *cp;
  for (u = di->firstUnredRel; u; u = di->firstUnredRel)
  {
    di->firstUnredRel = u->next;
    doubleFreeUnred(u);
  }
  for (rel = di->firstRedRel; rel; rel = di->firstRedRel)
  {
    di->firstRedRel = rel->next;
    doubleFreeRel(rel);
  }
  for (cp = di->firstCritPair; cp; cp = di->firstCritPair)
  {
    di->firstCritPair = cp->next;
    doubleFreeCritPair(cp);
  }
  if (di->secondSummand) freeHomogIdeal(di->secondSummand);
  if (di->benef) freeBenef(di->benef);
  free(di);
  return;
}

/******************************************************************************/
static void doubleUnlinkRedRel(dideal_t *di, drel_t *r)
{
  if (r->prev)
    r->prev->next = r->next;
  else
    di->firstRedRel = r->next;
  if (r->next) r->next->prev = r->prev;
  r->next = NULL;
  r->prev = NULL;
  return;
}

/******************************************************************************/
static void doubleInsertRedRelAfter(dideal_t *di, drel_t *prev, drel_t *r)
{
  drel_t *next = (prev) ? prev->next : di->firstRedRel;
  if (prev) prev->next = r; else di->firstRedRel = r;
  if (next) next->prev = r;
  r->prev = prev;
  r->next = next;
  return;
}

/******************************************************************************/
static void doubleInsertRedRel(dideal_t *di, drel_t *r)
{
  register drel_t *next, *prev = NULL;
  for (next = di->firstRedRel; next && leqTerm(r->tip, next->tip);
    next = next->next) prev = next;
  doubleInsertRedRelAfter(di, prev, r);
  return;
}

/******************************************************************************/
static void doubleUnlinkUnredRel(dideal_t *di, dunred_t *u)
{
  if (u->prev)
    u->prev->next = u->next;
  else
    di->firstUnredRel = u->next;
  if (u->next) u->next->prev = u->prev;
  u->next = NULL;
  u->prev = NULL;
  return;
}

/******************************************************************************/
static void doubleUnlinkCritPair(dideal_t *di, dpair_t *cp)
{
  if (cp->prev)
    cp->prev->next = cp->next;
  else
    di->firstCritPair = cp->next;
  if (cp->next) cp->next->prev = cp->prev;
  cp->next = NULL;
  cp->prev = NULL;
  return;
}

/******************************************************************************/
static void doubleInsertCritPair(dideal_t *di, dpair_t *cp)
{
  register dpair_t *next, *prev = NULL;
  for (next = di->firstCritPair; next && leqTerm(cp->lcm, next->lcm);
    next = next->next) prev = next;
  cp->prev = prev;
  cp->next = next;
  if (prev)
    prev->next = cp;
  else
    di->firstCritPair = cp;
  if (next) next->prev = cp;
  return;
}

/******************************************************************************/
static dpair_t *doubleGcdRel(freeGcr_t *fGcr, drel_t *r1, drel_t *r2)
{
  register dpair_t *dp;
  register cpair_t *cp = innerGcdRel(fGcr, r1->tip, r2->tip, false);   
  /* Always returns lcm/gcd, even if gcd=1 */
  dp = doubleNewCritPair();
  dp->lcm = cp->lcm; cp->lcm = NULL;
  dp->t1 = cp->t1; cp->t1 = NULL;
  dp->t2 = cp->t2; cp->t2 = NULL;
  dp->sign = cp->sign;
  dp->r1 = r1; dp->r2 = r2;
  freeCritPair(cp);
  return dp;
}

/******************************************************************************/
static void doubleRecordNewCritPairs(dideal_t *di, drel_t *r2)
{
  freeGcr_t *fGcr = di->fGcr;
  drel_t *r1;
  dpair_t *cp;
  for (r1 = di->firstRedRel; r1; r1 = r1->next)
  {
    cp = doubleGcdRel(fGcr, r1, r2);
    if (cp) doubleInsertCritPair(di, cp);
  }
  return;
}

/******************************************************************************/
static void doubleDiscardTheseCritPairs(dideal_t *di, drel_t *r)
{
  dpair_t *cp, *next;
  for (cp = di->firstCritPair; cp; cp = next)
  {
    next = cp->next;
    if (cp->r1 == r || cp->r2 == r)
    {
      doubleUnlinkCritPair(di, cp);
      doubleFreeCritPair(cp);
    }
  }
  return;
}

/******************************************************************************/
static drel_t *doubleReducingRel(dideal_t *di, cterm_t *t)
/* Returns a rel whose tip divides t, or NULL if none such */
{
  register drel_t *r;
  for (r = di->firstRedRel; r; r = r->next)
    if (isDivisible(t, r->tip)) return r;
  return NULL;
}

/******************************************************************************/
static dclass_t *doubleProductClassTerm(freeGcr_t *fGcr, dclass_t *d1,
  cterm_t *t2)
{
  cclass_t *c1 = productClassTerm(fGcr, d1->c[0], t2);
  cclass_t *c2 = productClassTerm(fGcr, d1->c[1], t2);
  dclass_t *d = doubleNewClass(c1, c2);
  return d;
}

/******************************************************************************/
static void doubleAddmulClass(dclass_t *d1, dclass_t *d2, FEL f)
/* sets d1 = d1 + f * d2 */
{
  addmulClass(d1->c[0], d2->c[0], f);
  addmulClass(d1->c[1], d2->c[1], f);
  return;
}

/******************************************************************************/
static void doubleReduceClassTillTipIrred(dideal_t *di, dclass_t *d,
  boolean decsOnly)
{
  freeGcr_t *fGcr = di->fGcr;
  drel_t *r;
  register cmon_t *m, *q;
  register dclass_t *valq;
  cclass_t *c = d->c[0];
  FEL f;
  while ((m = c->firstMon) && (r = doubleReducingRel(di, m->term)))
  {
    if (decsOnly && !r->decom && r->tip->dim == m->term->dim)
      return;
    q = rightFactor(fGcr, m->term, r->tip);
    f = zmul(m->coeff, q->coeff);
    valq = doubleProductClassTerm(fGcr, r->val, q->term);
    freeMon(q);
    unlinkAndDiscardMon(c, m);
    doubleAddmulClass(d, valq, f);
    doubleFreeClass(valq);
  }
  return;
}

/******************************************************************************/
static void doubleReduceUnredTillTipIrred(dideal_t *di, dunred_t *u)
{
  doubleReduceClassTillTipIrred(di, u->d, u->decom);
}

/******************************************************************************/
static void doubleCompletelyReduceClassTail(dideal_t *di, dclass_t *d,
  boolean decsOnly)
/* Assumes d->c[0]->firstMon exists and is irreducible */
{
  freeGcr_t *fGcr = di->fGcr;
  cclass_t *c = d->c[0];
  register drel_t *r;
  register cmon_t *m, *q, *prev = c->firstMon;
  dclass_t *valq;
  FEL f;
  if (!prev) OtherError("doubleCompletelyReduceClassTail: zero class");
  while ((m = prev->next))
  {
    r = doubleReducingRel(di, m->term);
    if (!r || (decsOnly && !r->decom && r->tip->dim == m->term->dim))
    {
      prev = m;
      continue;
    }
    q = rightFactor(fGcr, m->term, r->tip);
    f = zmul(m->coeff, q->coeff);
    valq = doubleProductClassTerm(fGcr, r->val, q->term);
    freeMon(q);
    unlinkAndDiscardMon(c, m);
    doubleAddmulClass(d, valq, f);
    doubleFreeClass(valq);
  }
  return;
}

/******************************************************************************/
static void doubleCompletelyReduceClass(dideal_t *di, dclass_t *d,
  boolean decsOnly)
{
  doubleReduceClassTillTipIrred(di, d, decsOnly);
  if (LTinFirstSummand(d)) doubleCompletelyReduceClassTail(di, d, decsOnly);
  return;
}

/******************************************************************************/
static void doubleMulClass(dclass_t *d, FEL f)
{
  mulClass(d->c[0], f);
  mulClass(d->c[1], f);
  return;
}

/******************************************************************************/
static drel_t *doublePromotedUnredRel(dunred_t *u)
/* Frees u but reuses u->d */
{
  FEL f;
  drel_t *r;
  dclass_t *d = u->d;
  cclass_t *c = d->c[0];
  cmon_t *m = c->firstMon;
  if (!m) OtherError("doublePromotedUnredRel: no leading term");
  f = zneg(zinv(m->coeff));
  unlinkMon(c, m);
  doubleMulClass(d, f);
  r = doubleNewRel(m->term, d, u->decom);
  m->term = NULL;
  freeMon(m);
  u->d = NULL;
  doubleFreeUnred(u);
  return r;
}

/******************************************************************************/
static dunred_t *doubleDemotedRedRel(drel_t *r)
/* Frees r */
{
  dunred_t *u;
  dclass_t *d = r->val;
  cclass_t *c = d->c[0];
  cmon_t *m = newMon(zneg(F_ONE), r->tip);
  insertMonAfter(c, NULL, m);
  u = doubleNewUnred(d, r->decom);
  r->tip = NULL;
  r->val = NULL;
  doubleFreeRel(r);
  return u;
}

/******************************************************************************/
static void doubleUnseatRedRel(dideal_t *di, drel_t *r)
{
  dunred_t *u;
  doubleUnlinkRedRel(di, r);
  doubleDiscardTheseCritPairs(di, r);
  u = doubleDemotedRedRel(r); /* Destroys + frees r */
  doubleInsertUnredRel(di, u);
  return;
}

/******************************************************************************/
void doubleProcessNewRedRel(dideal_t *di, drel_t *new)
{
  register drel_t *r, *old;
  for (old = di->firstRedRel; old; old = r)
  {
    r = old->next;
    if (isDivisible(old->tip, new->tip))
      doubleUnseatRedRel(di, old);
  }
  doubleRecordNewCritPairs(di, new);
  doubleInsertRedRel(di, new);
  for (r = di->firstRedRel; r; r = r->next)
    doubleCompletelyReduceClass(di, r->val, r->decom);
  return;
}

/******************************************************************************/
static void doubleFileAwayReducedUnred(dideal_t *di, dunred_t *u)
{
  if (LTinFirstSummand(u->d))
  {
    drel_t *r;
    r = doublePromotedUnredRel(u);
    doubleProcessNewRedRel(di, r);
    return;
  }
  if (LTinSecondSummand(u->d))
  {
    cunred_t *uu;
    uu = newUnred(u->d->c[1], false);
    insertUnredRel(di->secondSummand, uu);
    u->d->c[1] = NULL;
    doubleFreeUnred(u);
    return;
  }
  /* Now d = 0 guaranteed */
  doubleFreeUnred(u);
  return;
}

/******************************************************************************/
static void doubleIncorporateOneLevel(dideal_t *di, long n)
{
  register dunred_t *u;
  if ((u = di->firstUnredRel) && doubleDimUnred(di, u) < n)
    OtherError("doubleIncorporateOneLevel: unred reln in unexpectedly low dim");
  while ((u = di->firstUnredRel) && doubleDimUnred(di, u) == n)
  {
    doubleUnlinkUnredRel(di, u);
    doubleReduceUnredTillTipIrred(di, u);
    doubleFileAwayReducedUnred(di, u);
  }
  return;
}

/******************************************************************************/
static dclass_t *doubleValueOfCritPair(freeGcr_t *fGcr, dpair_t *cp)
{
  dclass_t *d1 = doubleProductClassTerm(fGcr, cp->r1->val, cp->t1);
  dclass_t *d2 = doubleProductClassTerm(fGcr, cp->r2->val, cp->t2);
  FEL f = zneg(cp->sign);
  doubleAddmulClass(d1, d2, f);
  doubleFreeClass(d2);
  return d1;
}

/******************************************************************************/
static void doubleFileAwayNewUnred(dideal_t *di, dclass_t *d)
{
  if (LTinFirstSummand(d))
  {
    dunred_t *u = doubleNewUnred(d, true);
    doubleInsertUnredRel(di, u);
    return;
  }
  if (LTinSecondSummand(d))
  {
    cunred_t *u = newUnred(d->c[1], false);
    d->c[1] = NULL;
    doubleFreeClass(d);
    insertUnredRel(di->secondSummand, u);
    return;
  }
  /* Now d = 0 guaranteed */
  doubleFreeClass(d);
  return;
}

/******************************************************************************/
static void doubleExpandCritPair(dideal_t *di, dpair_t *cp)
{
  dclass_t *d = doubleValueOfCritPair(di->fGcr, cp);
  doubleFileAwayNewUnred(di, d);
  doubleUnlinkCritPair(di, cp);
  doubleFreeCritPair(cp);
  return;
}

/******************************************************************************/
static void doubleExpandOneLevel(dideal_t *di, long n)
{
  register dpair_t *cp, *next;
  if (di->firstCritPair && di->firstCritPair->lcm->dim < n)
    OtherError("doubleExpandOneLevel: crit pair in unexpectedly low dimension");
  for (cp = di->firstCritPair; cp && cp->lcm->dim == n; cp = next)
  {
    next = cp->next;
    doubleExpandCritPair(di, cp);
  }
  return;
}

/******************************************************************************/
void doubleBuchbergerOnThisLevel(dideal_t *di, long n)
{
  long n0;
  doubleReactToAnyNewExteriorGenerators(di);
  n0 = doubleSmallestInterestingDimension(di);
  if (n0 == -1 || n0 > n) return;
  if (n0 < n) OtherError("doubleBuchbergerOnThisLevel: illegal skip");
  doubleExpandOneLevel(di, n);
  doubleIncorporateOneLevel(di, n);
  if (n - di->diff >= 0)
    buchbergerOnThisLevel(di->secondSummand, n - di->diff);
  return;
}

/******************************************************************************/
void doubleBuchberger(dideal_t *di)
{
  register long n;
  while ((n = doubleSmallestInterestingDimension(di)) != -1)
  {
    doubleBuchbergerOnThisLevel(di, n);
  }
  buchberger(di->secondSummand);
  return;
}

/******************************************************************************/
hideal_t *idealIntersection(hideal_t *hi1, hideal_t *hi2)
{
  dideal_t *di;
  hideal_t *hi;
  freeGcr_t *fGcr;
  register crel_t *r;
  register cclass_t *c1, *c2;
  register dclass_t *d;
  register dunred_t *u;
  fGcr = hi1->fGcr;
  if (hi2->fGcr != fGcr) OtherError("idealIntersection: mismatch");
  di = doubleNewHomogIdeal(fGcr, 0);
  hi = di->secondSummand;
  for (r = hi1->firstRedRel; r; r = r->next)
  {
    if (r->decom) continue;
    c1 = relClass(r);
    c2 = relClass(r);
    d = doubleNewClass(c1, c2);
    u = doubleNewUnred(d, false);
    doubleInsertUnredRel(di, u);
  }
  for (r = hi2->firstRedRel; r; r = r->next)
  {
    if (r->decom) continue;
    c1 = relClass(r);
    c2 = newClass();
    d = doubleNewClass(c1, c2);
    u = doubleNewUnred(d, false);
    doubleInsertUnredRel(di, u);
  }
  doubleBuchberger(di);
  di->secondSummand = NULL;
  doubleFreeHomogIdeal(di);
  return hi;
}

/******************************************************************************/
hideal_t *annihilatorIdeal(hideal_t *relns, cclass_t *cin)
{
  cclass_t *c;
  freeGcr_t *fGcr;
  hideal_t *ann = NULL;
  if (smallestInterestingDimension(relns) != -1)
    OtherError("annihilator ideal: Groebner basis for relns not complete");
  fGcr = relns->fGcr;
  c = reducedDuplicateClass(relns, cin);
  if (!c->firstMon)
  {
    ann = improperIdeal(fGcr);
  }
  else
  {
    long diff;
    dideal_t *di;
    register cclass_t *c1, *c2;
    register dclass_t *d;
    register dunred_t *u;
    register crel_t *r;
    diff = c->firstMon->term->dim;
    di = doubleNewHomogIdeal(fGcr, diff);
    c1 = duplicateClass(c);
    c2 = oneClass(fGcr);
    d = doubleNewClass(c1,c2);
    u = doubleNewUnred(d, false);
    doubleInsertUnredRel(di, u);
    for (r = relns->firstRedRel; r; r = r->next)
    {
      if (r->decom) continue;
      c1 = relClass(r);
      c2 = newClass();
      d = doubleNewClass(c1,c2);
      u = doubleNewUnred(d, false);
      doubleInsertUnredRel(di, u);
    }
    doubleBuchberger(di);
    ann = di->secondSummand;
    di->secondSummand = NULL;
    doubleFreeHomogIdeal(di);
  }
  freeClass(c);
  #ifdef DJG_DEBUG
  {
    boolean OK = leqIdeal(relns, ann);
    if (!OK) OtherError("annihilatorIdeal: annihilator should contain relns");
  }
  #endif
  return ann;
}

/******************************************************************************/
hideal_t *sequenceAnnihilator(hideal_t *relns, cclass_t **seq, long len)
{
  hideal_t *hi;
  if (len <= 0) OtherError("sequenceAnnihilator: len must be >= 1");
  if (len == 1)
    hi = annihilatorIdeal(relns, seq[0]);
  else
  {
    hideal_t **ann, **cap;
    register long i;
    ann = (hideal_t **) malloc(len << PTRSH);
    cap = (hideal_t **) malloc((len - 1) << PTRSH);
    if (!ann || !cap) AllocationError("sequenceAnnihilator");
    for (i = 0; i < len; i++)
    {
      ann[i] = annihilatorIdeal(relns, seq[i]);
      if (i == 1)
      {
        cap[0] = idealIntersection(ann[0], ann[1]);
        freeHomogIdeal(ann[0]);
        freeHomogIdeal(ann[1]);
      }
      if (i >= 2)
      {
        cap[i-1] = idealIntersection(cap[i-2], ann[i]);
        freeHomogIdeal(cap[i-2]);
        freeHomogIdeal(ann[i]);
      }
    }
    hi = cap[len-2];
    free(ann);
    free(cap);
  }
  return hi;
}

/******************************************************************************/
boolean isRegularElement(hideal_t *relns, cclass_t *cin)
{
  cclass_t *c;
  boolean result = false;
  c = reducedDuplicateClass(relns, cin);
  if (c->firstMon)
  {
    hideal_t *ann;
    ann = annihilatorIdeal(relns, c);
    result = leqIdeal(ann, relns);
    freeHomogIdeal(ann);
  }
  freeClass(c);
  return result;
}

/******************************************************************************/
boolean isRegularSequence(hideal_t *relns, cclass_t **seq, long len)
{
  register hideal_t *hi;
  register long i;
  register cclass_t *c;
  register cunred_t *u;
  boolean result = true;
  hi = duplicateGroebnerHomogIdeal(relns);
  for (i = 0; i < len; i++)
  {
    c = duplicateClass(seq[i]);
    reduceClassTillTipIrred(hi, c, false);
    if (!c->firstMon || !isRegularElement(hi, c))
    {
      freeClass(c);
      result = false;
      break;
    }
    u = newUnred(c, false);
    insertUnredRel(hi, u);
    buchberger(hi);
  }
  freeHomogIdeal(hi);
  return result;
}
