/* ========================== C MeatAxe =============================
   Basic row operations.

   (C) Copyright 1994 Michael Ringe, Lehrstuhl D fuer Mathematik,
   RWTH Aachen, Germany  <mringe@tiffy.math.rwth-aachen.de>
   This program is free software; see the file COPYING for details.
   ================================================================== */


/* $Id: zzz2.c,v 1.2 1997/09/11 15:44:43 gap Exp $
 *
 * $Log: zzz2.c,v $
 * Revision 1.2  1997/09/11 15:44:43  gap
 * New version 2.2.3. AH
 *
 * Revision 1.4  1994/11/25  14:08:17  mringe
 * ANSI-C, neue Namen fuer time-Funktionen.
 *
 * Revision 1.3  1994/07/29  07:41:52  mringe
 * Benutze zmulrow() zum Initialisieren in zalloc().
 *
 * Revision 1.2  1994/07/28  06:04:43  mringe
 * zsetfield() und zsetlen() als getrennte Funktionen.
 *
 * Revision 1.1  1994/07/25  18:33:53  mringe
 * Initial revision
 *
 * Revision 1.1  1994/07/20  22:19:39  mringe
 * Initial revision
 *
 *
 */


#include <string.h>
#include <stdlib.h>
#include "meataxe.h"

#define LPR (zrowsize/sizeof(long))

/* ------------------------------------------------------------------
   zalloc()  -  Allocate memory and initialize
   ------------------------------------------------------------------ */

PTR zalloc(long nrows)

{
    PTR p;
    char *q;
    register long i;
    size_t req = (size_t)(zrowsize*nrows);

    if ((p = (PTR) Smalloc(req)) == NULL)
    {
	fprintf(stderr,"zalloc(): Request=%ld Bytes\n",(long)req);
	FATAL("OUT OF MEMORY");
    }

    /* Initialize */
    q = (char *) p;
    for (i = nrows; i > 0; --i)
    {
	zmulrow((PTR) q,F_ZERO);
	q += zrowsize;
    }
    return p;
}


/* ------------------------------------------------------------------
   zfree()  -  Deallocate memory 
   ------------------------------------------------------------------ */

inline void zfree(PTR p) { free(p); return; }
/* Simon King (2008-12) defining it inline */

/* ------------------------------------------------------------------
   zsize()  -  Size of a matrix or permutation(s)
   ------------------------------------------------------------------ */

inline size_t zsize(long nrows) { return ((size_t)(zrowsize * nrows)); }
/* Simon King (2008-12) defining it inline */


/* ------------------------------------------------------------------
   zmoverow()  -  Move row/vector

   return value: none
   ------------------------------------------------------------------ */

inline void zmoverow(PTR dest, PTR src) { memcpy(dest,src,zrowsize_io); } 
//zsize((long) 1)); }
/* Simon King (2008-12) defining it inline
{
  memcpy(dest,src,zsize((long) 1));
}
*/

/* This is the original version:
  {
    register long *s = (long *) src, *d = (long *) dest, i;
    for (i = LPR; i > 0; --i)
   	 *d++ = *s++;
}
*/

/* ------------------------------------------------------------------
   Added by Simon King (2008-03-31)
   zcpyrow()  -  copy memory block of a certain length

   return value: none
   ------------------------------------------------------------------ */

void zcpyrow(PTR dest, PTR src, long sz) { memcpy(dest,src,sz); }
// dest must be allocated enough memory!
/* Simon King (2008-12) defining it inline
{
  memcpy(dest,src,sz);
}
*/

/* ------------------------------------------------------------------
   zswaprow()  -  Swap two rows (modified by Simon King)

   return value: none
   ------------------------------------------------------------------ */

void zswaprow(PTR dest, PTR src)

{ register PTR tmp = zalloc(1);
  if (tmp == NULL)
	MTXFAIL(ERR_NOMEM,NULL);
  zmoverow(tmp, dest);
  zmoverow(dest, src);
  zmoverow(src, tmp);
  zfree(tmp);
}

/* original version
void zswaprow(PTR dest, PTR src)

{	register long *p1 = (long*) src, *p2 = (long*) dest, l;

	for (i = LPR; i > 0; --i)
	{	l = *p1;
		*p1++ = *p2;
		*p2++ = l;
	}

}
*/

/* ------------------------------------------------------------------
   zcmprow()  -  Compare two rows (modified by Simon King
   ------------------------------------------------------------------ */

int zcmprow(PTR dest, PTR src)

{	
    register long k;
    register long long*d = (long long*) dest, *s = (long long*) src;
    register size_t LLPR = (LPR << LONGSH);
    LLPR = LLPR / sizeof(long long);

    for (k = 0; k < LLPR; ++k)
	if (*d++ != *s++) return 1;
    k = (k * sizeof(long long)) >> LONGSH;
    register long *d2 = (long *)d, *s2 = (long *)s;
    for (; k < LPR; k++)
        if (*d2++ != *s2++) return 1;
    return 0;
}

/*  original version
int zcmprow(PTR dest, PTR src)

{
    register long k;
    register long *d = (long*) dest, *s = (long*) src;

    for (k = LPR; k > 0; --k)
        if (*d++ != *s++) return 1;
    return 0;
}
*/


/* ------------------------------------------------------------------
 *   zadvance()  -  advance pointer
 *
 *  return value: none
 *  Simon King (2008-12) defining it as macro. 
 *  ------------------------------------------------------------------ 
 *inline void zadvance(PTR *ptr, long nrows) { *(long **)ptr += nrows * LPR; }
*/

/* ------------------------------------------------------------------
 * Two new functions, eventually also defined as macros
 *
 *zsteprow()  -  move pointer one row forwards  (Simon King 2008-12)
 *
 * return value: none
 * ------------------------------------------------------------------- 
 *inline void zsteprow(PTR *ptr) { *(long **)ptr += LPR; }
 *
 *------------------------------------------------------------------
 * zbackrow()  -  move pointer one row backwards  (Simon King 2008-12)
 *
 * return value: none
 * ------------------------------------------------------------------- 
 *inline void zbackrow(PTR *ptr) { *(long **)ptr -= LPR; }
*/

/* ------------------------------------------------------------------
   zpermrow()  -  multiply a row by a permutation
   ------------------------------------------------------------------ */

PTR zpermrow(PTR row, long *perm, PTR result)
{
    register FEL f;
    register long i;
    register long *p = (long *)perm;
    PTR byteptr = row;
    int idx=0;

    for (i = 1; i <= znoc; ++i)
    {
	f = zextract_step(&byteptr,&idx); /*row,i);*/
	zinsert(result,*p++,f);
    }
    return result;
}



/* ------------------------------------------------------------------
   mtxinit()  -  Initialize
   ------------------------------------------------------------------ */

int mtxinit()

{
    zchar = 0;
    zgen = 0;
    zfl = 0;		/* This forces zsetlen() to read the tables! */
    SInit();
    return (ZZZVERSION);
}

